#include "stdafx.h"
#include "HDUSTest.h"
#include "ProgramListView.h"


#define PROGRAM_LIST_WINDOW_CLASS TEXT("HDUSTest Program List")

#define TEXT_LEFT_MARGIN 8



class CProgramItemInfo {
	EVENT_INFO_DATA m_EventInfo;
	WORD m_EventID;
	int m_NameLines;
	int m_TextLines;
	int CalcLines(HDC hdc,LPCTSTR pszText,int Width);
public:
	CProgramItemInfo(EVENT_INFO_DATA *pEventInfo);
	//~CProgramItemInfo();
	const EVENT_INFO_DATA &GetEventInfo() const { return m_EventInfo; }
	WORD GetEventID() const { return m_EventID; }
	int GetTitleLines() const { return m_NameLines; }
	int GetTextLines() const { return m_TextLines; }
	int GetLines() const { return m_NameLines+m_TextLines; }
	int CalcLines(HDC hdc,int Width);
	void DrawString(HDC hdc,LPCTSTR pszText,const RECT *pRect,int LineHeight);
	void DrawTitle(HDC hdc,const RECT *pRect,int LineHeight);
	void DrawText(HDC hdc,const RECT *pRect,int LineHeight);
};


CProgramItemInfo::CProgramItemInfo(EVENT_INFO_DATA *pEventInfo)
{
	m_EventInfo=*pEventInfo;
	m_EventID=pEventInfo->wEventID;
	m_NameLines=0;
	m_TextLines=0;
}


int CProgramItemInfo::CalcLines(HDC hdc,LPCTSTR pszText,int Width)
{
	LPCTSTR p;
	int Length;
	int Fit;
	SIZE sz;
	int Lines;

	Lines=0;
	p=pszText;
	while (*p!='\0') {
		if (*p=='\r' || *p=='\n') {
			p++;
			if (*p=='\n')
				p++;
			if (*p=='\0')
				break;
			Lines++;
			continue;
		}
		for (Length=0;p[Length]!='\0' && p[Length]!='\r' && p[Length]!='\n';Length++);
		GetTextExtentExPoint(hdc,p,Length,Width,&Fit,NULL,&sz);
		if (Fit<1)
			Fit=1;
		p+=Fit;
		Lines++;
	}
	return Lines;
}


int CProgramItemInfo::CalcLines(HDC hdc,int Width)
{
	RECT rc;
	TCHAR szText[2048];

	wsprintf(szText,TEXT("%02d:%02d"),m_EventInfo.stStartTime.wHour,
											m_EventInfo.stStartTime.wMinute);
	if (!m_EventInfo.strEventName.IsEmpty()) {
		wsprintf(szText+lstrlen(szText),TEXT(" %s"),
										m_EventInfo.strEventName.GetBuffer(0));
	}
	m_NameLines=CalcLines(hdc,szText,Width);
	if (!m_EventInfo.strEventText.IsEmpty())
		m_TextLines=CalcLines(hdc,m_EventInfo.strEventText.GetBuffer(0),Width-TEXT_LEFT_MARGIN);
	else
		m_TextLines=0;
	return m_NameLines+m_TextLines;
}


void CProgramItemInfo::DrawString(HDC hdc,LPCTSTR pszText,const RECT *pRect,
																int LineHeight)
{
	LPCTSTR p;
	int y;
	int Length;
	int Fit;
	SIZE sz;
	int Lines;

	p=pszText;
	y=pRect->top;
	while (*p!='\0') {
		if (*p=='\r' || *p=='\n') {
			p++;
			if (*p=='\n')
				p++;
			y+=LineHeight;
			continue;
		}
		for (Length=0;p[Length]!='\0' && p[Length]!='\r' && p[Length]!='\n';Length++);
		GetTextExtentExPoint(hdc,p,Length,pRect->right-pRect->left,&Fit,NULL,&sz);
		if (Fit<1)
			Fit=1;
		::TextOut(hdc,pRect->left,y,p,Fit);
		p+=Fit;
		y+=LineHeight;
	}
}


void CProgramItemInfo::DrawTitle(HDC hdc,const RECT *pRect,int LineHeight)
{
	TCHAR szText[2048];

	wsprintf(szText,TEXT("%02d:%02d"),m_EventInfo.stStartTime.wHour,
											m_EventInfo.stStartTime.wMinute);
	if (!m_EventInfo.strEventName.IsEmpty()) {
		wsprintf(szText+lstrlen(szText),TEXT(" %s"),
										m_EventInfo.strEventName.GetBuffer(0));
	}
	DrawString(hdc,szText,pRect,LineHeight);
}


void CProgramItemInfo::DrawText(HDC hdc,const RECT *pRect,int LineHeight)
{
	if (!m_EventInfo.strEventText.IsEmpty()) {
		DrawString(hdc,m_EventInfo.strEventText.GetBuffer(0),pRect,LineHeight);
	}
}




HINSTANCE CProgramListView::m_hinst=NULL;


bool CProgramListView::Initialize(HINSTANCE hinst)
{
	WNDCLASS wc;

	m_hinst=hinst;
	wc.style=CS_HREDRAW;
	wc.lpfnWndProc=WndProc;
	wc.cbClsExtra=0;
	wc.cbWndExtra=0;
	wc.hInstance=hinst;
	wc.hIcon=NULL;
	wc.hCursor=LoadCursor(NULL,IDC_ARROW);
	wc.hbrBackground=NULL;
	wc.lpszMenuName=NULL;
	wc.lpszClassName=PROGRAM_LIST_WINDOW_CLASS;
	return RegisterClass(&wc)!=0;
}


CProgramListView::CProgramListView()
{
	LOGFONT lf;

	m_pProgramList=NULL;
	m_hwnd=NULL;
	GetObject(GetStockObject(DEFAULT_GUI_FONT),sizeof(LOGFONT),&lf);
	m_hfont=CreateFontIndirect(&lf);
	m_FontHeight=abs(lf.lfHeight);
	m_LineMargin=1;
	m_crBackColor=RGB(0,0,0);
	m_crTextColor=RGB(255,255,255);
	m_crTitleBackColor=RGB(128,128,128);
	m_crTitleTextColor=RGB(255,255,255);
	m_ppItemList=NULL;
	m_NumInfoListItems=0;
	m_ScrollPos=0;
}


CProgramListView::~CProgramListView()
{
	DeleteObject(m_hfont);
	if (m_ppItemList!=NULL) {
		for (int i=0;i<m_NumInfoListItems;i++)
			delete m_ppItemList[i];
		delete [] m_ppItemList;
	}
}


bool CProgramListView::Create(HWND hwndParent,DWORD Style,DWORD ExStyle)
{
	if (m_hwnd!=NULL)
		return true;
	m_hwnd=CreateWindowEx(ExStyle,PROGRAM_LIST_WINDOW_CLASS,TEXT("ԑg\"),Style,
		m_WindowPosition.Left,m_WindowPosition.Top,
		m_WindowPosition.Width,m_WindowPosition.Height,
		hwndParent,NULL,m_hinst,this);
	return m_hwnd!=NULL;
}


bool CProgramListView::UpdateProgramList(WORD ServiceID)
{
	if (m_pProgramList==NULL)
		return false;
	m_pProgramList->UpdateProgramList();
	if (m_hwnd!=NULL) {
		if (UpdateListInfo(ServiceID)) {
			CalcDimentions();
			SetScrollBar();
			InvalidateRect(m_hwnd,NULL,TRUE);
		}
	}
	return true;
}


bool CProgramListView::UpdateListInfo(WORD ServiceID)
{
	CEpgProgramList::ServiceIterator itrService;
	CEpgProgramList::EventListIterator itrEvent;
	CEpgProgramList::EventIterator itrData;
	int NumEvents;
	int i;
	CProgramItemInfo **ppItemList;
	SYSTEMTIME st;
	FILETIME ft,ftPg;
	bool fChanged;

	if (m_pProgramList==NULL)
		return false;
	itrService=m_pProgramList->ServiceMap.begin();
	if (itrService==m_pProgramList->ServiceMap.end())
		return false;
	while (true) {
		if (itrService->second.wServiceID==ServiceID)
			break;
		itrService++;
		if (itrService==m_pProgramList->ServiceMap.end()) {
			itrService=m_pProgramList->ServiceMap.begin();
			break;
		}
	}
	itrEvent=m_pProgramList->EventMap.find(itrService->first);
	if (itrEvent==m_pProgramList->EventMap.end())
		return false;
	NumEvents=itrEvent->second.EventDataMap.size();
	if (NumEvents==0) {
		ClearItemList();
		return true;
	}
	ppItemList=new CProgramItemInfo*[NumEvents];
	GetLocalTime(&st);
	SystemTimeToFileTime(&st,&ft);
	ft+=(LONGLONG)24*60*60*10000000;
	fChanged=false;
	for (itrData=itrEvent->second.EventDataMap.begin(),i=0;itrData!=itrEvent->second.EventDataMap.end();itrData++,i++) {
		SystemTimeToFileTime(&itrData->second.stStartTime,&ftPg);
		if (CompareFileTime(&ft,&ftPg)<0)
			break;
		ppItemList[i]=new CProgramItemInfo(&itrData->second);
		if (i>=m_NumInfoListItems
				|| itrData->second.wEventID!=m_ppItemList[i]->GetEventID())
			fChanged=true;
	}
	if (i==m_NumInfoListItems && !fChanged) {
		for (i--;i>=0;i--)
			delete ppItemList[i];
		delete [] ppItemList;
		return false;
	}
	ClearItemList();
	m_ppItemList=ppItemList;
	m_NumInfoListItems=i;
	return true;
}


void CProgramListView::ClearItemList()
{
	if (m_ppItemList!=NULL) {
		int i;

		for (i=0;i<m_NumInfoListItems;i++)
			delete m_ppItemList[i];
		delete [] m_ppItemList;
		m_ppItemList=NULL;
		m_NumInfoListItems=0;
	}
}


void CProgramListView::ClearProgramList()
{
	if (m_NumInfoListItems>0) {
		ClearItemList();
		m_ScrollPos=0;
		m_TotalLines=0;
		if (m_hwnd!=NULL) {
			SetScrollBar();
			InvalidateRect(m_hwnd,NULL,TRUE);
		}
	}
	if (m_pProgramList!=NULL)
		m_pProgramList->Clear();
}


void CProgramListView::CalcDimentions()
{
	HDC hdc;
	RECT rc;
	HFONT hfontOld;
	int i;

	hdc=GetDC(m_hwnd);
	GetClientRect(m_hwnd,&rc);
	hfontOld=SelectFont(hdc,m_hfont);
	m_TotalLines=0;
	for (i=0;i<m_NumInfoListItems;i++) {
		m_TotalLines+=m_ppItemList[i]->CalcLines(hdc,rc.right);
	}
	SelectFont(hdc,hfontOld);
	ReleaseDC(m_hwnd,hdc);
}


void CProgramListView::SetScrollBar()
{
	SCROLLINFO si;
	RECT rc;

	si.cbSize=sizeof(SCROLLINFO);
	si.fMask=SIF_PAGE | SIF_RANGE | SIF_POS | SIF_DISABLENOSCROLL;
	si.nMin=0;
	si.nMax=m_TotalLines<1?0:m_TotalLines-1;
	GetClientRect(m_hwnd,&rc);
	si.nPage=rc.bottom/(m_FontHeight+m_LineMargin);
	si.nPos=m_ScrollPos;
	SetScrollInfo(m_hwnd,SB_VERT,&si,TRUE);
}


void CProgramListView::SetColors(COLORREF crBackColor,COLORREF crTextColor,
						COLORREF crTitleBackColor,COLORREF crTitleTextColor)
{
	m_crBackColor=crBackColor;
	m_crTextColor=crTextColor;
	m_crTitleBackColor=crTitleBackColor;
	m_crTitleTextColor=crTitleTextColor;
	if (m_hwnd!=NULL)
		InvalidateRect(m_hwnd,NULL,TRUE);
}


CProgramListView *CProgramListView::GetThis(HWND hwnd)
{
	return reinterpret_cast<CProgramListView*>(GetWindowLongPtr(hwnd,GWLP_USERDATA));
}


LRESULT CALLBACK CProgramListView::WndProc(HWND hwnd,UINT uMsg,WPARAM wParam,LPARAM lParam)
{
	switch (uMsg) {
	case WM_CREATE:
		{
			CProgramListView *pThis=static_cast<CProgramListView*>(
					reinterpret_cast<LPCREATESTRUCT>(lParam)->lpCreateParams);

			SetWindowLongPtr(hwnd,GWLP_USERDATA,
											reinterpret_cast<LONG_PTR>(pThis));
			pThis->m_hwnd=hwnd;
		}
		return 0;

	case WM_PAINT:
		{
			CProgramListView *pThis=GetThis(hwnd);
			PAINTSTRUCT ps;

			BeginPaint(hwnd,&ps);
			pThis->DrawProgramList(ps.hdc,&ps.rcPaint);
			EndPaint(hwnd,&ps);
		}
		return 0;

	case WM_SIZE:
		{
			CProgramListView *pThis=GetThis(hwnd);

			pThis->CalcDimentions();
			pThis->SetScrollBar();
		}
		return 0;

	case WM_MOUSEWHEEL:
		if (GET_WHEEL_DELTA_WPARAM(wParam)<0)
			wParam=SB_LINEDOWN;
		else
			wParam=SB_LINEUP;
	case WM_VSCROLL:
		{
			CProgramListView *pThis=GetThis(hwnd);
			int Pos;
			RECT rc;
			int Page;

			Pos=pThis->m_ScrollPos;
			GetClientRect(hwnd,&rc);
			Page=rc.bottom/(pThis->m_FontHeight+pThis->m_LineMargin);
			switch (LOWORD(wParam)) {
			case SB_LINEUP:		Pos--;				break;
			case SB_LINEDOWN:	Pos++;				break;
			case SB_PAGEUP:		Pos-=Page;			break;
			case SB_PAGEDOWN:	Pos+=Page;			break;
			case SB_THUMBTRACK:	Pos=HIWORD(wParam);	break;
			case SB_TOP:		Pos=0;				break;
			case SB_BOTTOM:		Pos=max(pThis->m_TotalLines-Page,0);	break;
			default:	return 0;
			}
			if (Pos<0)
				Pos=0;
			else if (Pos>max(pThis->m_TotalLines-Page,0))
				Pos=max(pThis->m_TotalLines-Page,0);
			if (Pos!=pThis->m_ScrollPos) {
				int Offset=Pos-pThis->m_ScrollPos;
				SCROLLINFO si;

				pThis->m_ScrollPos=Pos;
				si.cbSize=sizeof(SCROLLINFO);
				si.fMask=SIF_POS;
				si.nPos=Pos;
				SetScrollInfo(hwnd,SB_VERT,&si,TRUE);
				if (abs(-Offset*(pThis->m_FontHeight+pThis->m_LineMargin))<rc.bottom) {
					ScrollWindowEx(hwnd,0,-Offset*(pThis->m_FontHeight+pThis->m_LineMargin),
								NULL,NULL,NULL,NULL,SW_ERASE | SW_INVALIDATE);
				} else
					InvalidateRect(hwnd,NULL,TRUE);
			}
		}
		return 0;

	case WM_LBUTTONDOWN:
		SetFocus(hwnd);
		return 0;

	case WM_DESTROY:
		{
			CProgramListView *pThis=GetThis(hwnd);

			pThis->OnDestroy();
		}
		return 0;
	}
	return DefWindowProc(hwnd,uMsg,wParam,lParam);
}


void CProgramListView::DrawProgramList(HDC hdc,const RECT *prcPaint)
{
	HBRUSH hbr;
	COLORREF cr;
	HFONT hfontOld;
	COLORREF crOldTextColor;
	int OldBkMode;
	RECT rc;
	int i;

	hbr=CreateSolidBrush(m_crBackColor);
	FillRect(hdc,prcPaint,hbr);
	DeleteObject(hbr);
	hbr=CreateSolidBrush(m_crTitleBackColor);
	hfontOld=SelectFont(hdc,m_hfont);
	crOldTextColor=GetTextColor(hdc);
	OldBkMode=SetBkMode(hdc,TRANSPARENT);
	GetClientRect(m_hwnd,&rc);
	rc.top=-(m_ScrollPos*(m_FontHeight+m_LineMargin));
	for (i=0;i<m_NumInfoListItems;i++) {
		rc.bottom=rc.top+m_ppItemList[i]->GetTitleLines()*(m_FontHeight+m_LineMargin);
		if (rc.bottom>prcPaint->top) {
			SetTextColor(hdc,m_crTitleTextColor);
			rc.left=0;
			FillRect(hdc,&rc,hbr);
			m_ppItemList[i]->DrawTitle(hdc,&rc,m_FontHeight+m_LineMargin);
		}
		rc.top=rc.bottom;
		rc.bottom=rc.top+m_ppItemList[i]->GetTextLines()*(m_FontHeight+m_LineMargin);
		if (rc.bottom>prcPaint->top) {
			SetTextColor(hdc,m_crTextColor);
			rc.left=TEXT_LEFT_MARGIN;
			m_ppItemList[i]->DrawText(hdc,&rc,m_FontHeight+m_LineMargin);
		}
		rc.top=rc.bottom;
		if (rc.top>=prcPaint->bottom)
			break;
	}
	SetTextColor(hdc,crOldTextColor);
	SetBkMode(hdc,OldBkMode);
	SelectFont(hdc,hfontOld);
	DeleteObject(hbr);
}
